import { Entity, EntityEquippableComponent, EquipmentSlot, ItemDurabilityComponent, ItemEnchantableComponent, Vector3 } from "@minecraft/server";
import { ARMOR_PROTECTION_VALUES } from "armors.constantes";
import { Bullet } from "entities/bullet";
import { TWeapon } from "instances/global";
import { distanceVector } from "utils/distance-between-vector3";

export abstract class DamageHandler {

    private static getDamageTakenWithProtection(armorProtection: number, ArmorToughness: number, damage: number, enchantLevel : number): number {
        const top: number = Math.max(armorProtection / 5, armorProtection - ((4 * damage * (1/enchantLevel)) / ArmorToughness + 8));
        return 1 - top / 25;
    }

    public static handleArmorDurability(entity: Entity, damage: number): void {
		const equipment = entity.getComponent('equippable') as EntityEquippableComponent;

		const damageInHeart = damage * 2;
		const durabilityDamage = damageInHeart / 4;

		for (const equipmentSlot in EquipmentSlot) {
			if (equipmentSlot === 'Mainhand' || 'Offhand') continue;

			const item = equipment.getEquipment(EquipmentSlot[equipmentSlot]);
			if (!item) continue;

			const itemDurability = item.getComponent('durability') as ItemDurabilityComponent;
			itemDurability.damage -= durabilityDamage;
			equipment.setEquipment(EquipmentSlot[equipmentSlot], item);
		}
	}

	public static getKnockbackResistance(hit: Entity): number {
		const equipment = hit.getComponent('equippable') as EntityEquippableComponent;

		if (!equipment) return 1;

		const headEquipment = equipment.getEquipment(EquipmentSlot.Head);
		const chestEquipment = equipment.getEquipment(EquipmentSlot.Chest);
		const legsEquipment = equipment.getEquipment(EquipmentSlot.Legs);
		const feetEquipment = equipment.getEquipment(EquipmentSlot.Feet);

		const helmetKnockbackResistance: number = ARMOR_PROTECTION_VALUES[headEquipment?.typeId]?.knockback_resistance || 0;
		const chestplateKnockbackResistance: number = ARMOR_PROTECTION_VALUES[chestEquipment?.typeId]?.knockback_resistance || 0;
		const leggingsKnockbackResistance: number = ARMOR_PROTECTION_VALUES[legsEquipment?.typeId]?.knockback_resistance || 0;
		const feetKnockbackResistance: number = ARMOR_PROTECTION_VALUES[feetEquipment?.typeId]?.knockback_resistance || 0;

		const total = helmetKnockbackResistance + chestplateKnockbackResistance + leggingsKnockbackResistance + feetKnockbackResistance;
		return 1 - (total * 0.07);
	}

	private static handleArmorMultiplier(hit: Entity, damage: number): number {
		const equipment = hit.getComponent('equippable') as EntityEquippableComponent;

		if (!equipment) return 1;

        const headEquipment = equipment.getEquipment(EquipmentSlot.Head);
        const chestEquipment = equipment.getEquipment(EquipmentSlot.Chest);
        const legsEquipment = equipment.getEquipment(EquipmentSlot.Legs);
        const feetEquipment = equipment.getEquipment(EquipmentSlot.Feet);

		const helmetProtection: number = ARMOR_PROTECTION_VALUES[headEquipment?.typeId]?.protection || 0;
		const chestplateProtection: number = ARMOR_PROTECTION_VALUES[chestEquipment?.typeId]?.protection || 0;
		const leggingsProtection: number = ARMOR_PROTECTION_VALUES[legsEquipment?.typeId]?.protection || 0;
		const feetProtection: number = ARMOR_PROTECTION_VALUES[feetEquipment?.typeId]?.protection || 0;

		const maxProtection: number = helmetProtection + chestplateProtection + leggingsProtection + feetProtection;

		const helmetToughness: number = ARMOR_PROTECTION_VALUES[headEquipment?.typeId]?.toughness || 0;
		const chestplateToughness: number = ARMOR_PROTECTION_VALUES[chestEquipment?.typeId]?.toughness || 0;
		const leggingsToughness: number = ARMOR_PROTECTION_VALUES[legsEquipment?.typeId]?.toughness || 0;
		const feetToughness: number = ARMOR_PROTECTION_VALUES[feetEquipment?.typeId]?.toughness || 0;

		const maxToughness: number = helmetToughness + chestplateToughness + leggingsToughness + feetToughness;

		const helmetProtectionEnchant: number = (headEquipment?.getComponent('enchantable') as ItemEnchantableComponent)?.getEnchantment('protection')?.level || 0
		const chestplateProhelmetProtectionEnchant: number = (chestEquipment?.getComponent('enchantable') as ItemEnchantableComponent)?.getEnchantment('protection')?.level || 0
		const leggingsProhelmetProtectionEnchant: number = (legsEquipment?.getComponent('enchantable') as ItemEnchantableComponent)?.getEnchantment('protection')?.level || 0
		const feetProhelmetProtectionEnchant: number = (feetEquipment?.getComponent('enchantable') as ItemEnchantableComponent)?.getEnchantment('protection')?.level || 0

		const helmetProProtectionEnchant: number = (headEquipment?.getComponent('enchantable') as ItemEnchantableComponent)?.getEnchantment('protection')?.level || 0
		const chestplateProjectilProhelmetProtectionEnchant: number = (chestEquipment?.getComponent('enchantable') as ItemEnchantableComponent)?.getEnchantment('protection')?.level || 0
		const leggingsProjectilProhelmetProtectionEnchant: number = (legsEquipment?.getComponent('enchantable') as ItemEnchantableComponent)?.getEnchantment('protection')?.level || 0
		const feetProjewssWithEnchant: number = helmetToughness + chestplateToughness + leggingsToughness + feetToughness;

        const totalProtection: number = feetProjewssWithEnchant + helmetProProtectionEnchant + chestplateProjectilProhelmetProtectionEnchant + leggingsProjectilProhelmetProtectionEnchant + helmetProtectionEnchant + chestplateProhelmetProtectionEnchant + leggingsProhelmetProtectionEnchant + feetProhelmetProtectionEnchant;

		return DamageHandler.getDamageTakenWithProtection(maxProtection, maxToughness, damage, totalProtection) || 1;
	}

    public static getDamages(weapon : TWeapon, hit : Entity, bullet : Bullet) : number
    {
		const damages = weapon.damage - (weapon.damageDropOff * Math.floor(distanceVector(bullet.spawnLocation, hit.location)))
		if (damages <= 0) return 0;
		const armor = this.handleArmorMultiplier(hit, damages);
		if (armor == 1) return (damages);
        return (damages * ((this.handleArmorMultiplier(hit, damages) * (weapon.armorPenetration))))
    }
}
